from dataclasses import dataclass
import json

import requests


@dataclass
class Player:
    """Class with attributes of a Runescape Player using Jagex's RS3 API

    Most of Runescape 3's API can be accessed at: http://runescape.wikia.com/wiki/Application_programming_interface

    ...

    Args:
        name (str): Name of the Player, gets set to case-sensitive name later if his RuneMetrics profile is not private.
        If you're also working with clans and the player's profile is private, name should be passed case-sensitively.
        Otherwise some clan attributes will not be available.

        runemetrics (bool): Whether or not Runemetrics player info should be set or not, slows down creating the object,
        but provides more attributes and methods.

    Attributes:
        name (str): Name of the Player, gets set to case-sensitive name later if his RuneMetrics profile is not private.

        metrics_info (str): Runemetrics Info in dictionary format about the player.

        exp (int or None): Total Exp of the player, or None if private_profile is True.

        combat_level (int or None): Total Combat Level of the player, or None if private_profile is True.

        total_level (int or None): Total Skill Level of the player, or None if private_profile is True.

        quests_not_started (int): Number of quests the player has not started.

        quests_started (int): Number of the quests the player has started.

        quests_complete (int): Number of quests the player has completed.

        private_profile (bool): Whether the player has a private RuneMetrics profile or not.

        exists (bool): Whether the player exists or not.

        details_status_code (int): The status code of the GET request to RS3 Player Details API

        runemetrics_status_code (int): The status code of the GET request to RS3 Player Runemetrics API

    Raises:
        ConnectionError
            If connecting to any of RS3's API related to players fails.
    """
    name: str
    runemetrics: bool = True

    metrics_info: dict = None
    exp: int = None
    combat_level: int = None
    total_level: int = None
    quests_not_started: int = None
    quests_started: int = None
    quests_complete: int = None
    skill_values: list = None
    details_status_code: int = None
    runemetrics_status_code: int = None
    private_profile: bool = True
    exists: bool = True

    def __post_init__(self) -> None:
        if not self.name:
            raise TypeError("Player name cannot be null.")

        self.info = self._dict_info()
        self.suffix = self.info['is_suffix']
        self.title = self.info['title']

        # TODO: The check below is a band-aid fix that marks self.exists as False even if 'runemetrics' parameter
        # is passed as False when creating Player object, a real way to check if a player exists needs to be added
        if not self._valid_username(self.name):
            self.exists = False

        if self.runemetrics:
            self.set_runemetrics_info()

        try:
            self.clan = self.info['clan']
        except KeyError:
            self.clan = None

    @staticmethod
    def _valid_username(username: str) -> bool:
        """
        Checks if a username is in the valid character limit set by Jagex
        and if it doesn't have any invalid characters in it.

        Args:
            username (str): The username to be checked if valid or not

        Returns:
            bool. True if the username is valid, False if it is not
        """
        if len(username) > 12:
            return False
        banned_chars = (
            '!', '@', '#', '$', '%', '¨', '&', '*', '(', ')', '-', '[', ']', ',', '.', '~', '^'
                                                                                            '{', '}', ':', ';', "'",
            '"', 'ç', 'Ç', '?', '/', '°', '|', '\\', '`', '=', '+', '¬'
                                                                    '¹', '²', '³', '£', '¢', 'ª', 'º')
        if any(char in username for char in banned_chars):
            return False
        return True

    def _raw_info(self) -> str:
        """
        Gets JsonP information on a player, in string format.

        Returns:
            str. JsonP information on player, converted to string.

        Raises:
            ConnectionError
                If connecting to the RS3 Runemetrics profile API was not possible.

        ------
        .. note::
            Has to be formatted correctly into Json or other formats to be worked with properly first.
        """

        info_url = (f"http://services.runescape.com/m=website-data/"
                    f"playerDetails.ws?names=%5B%22{self.name}%22%5D&callback=jQuery000000000000000_0000000000&_=0")
        response = requests.get(info_url)
        self.details_status_code = response.status_code
        if response.status_code != 200:
            raise ConnectionError(f'Not able to connect to RS3 playerDetails API. Status code: {response.status_code}')
        return str(response.content)

    def _dict_info(self) -> dict:
        """
        Gets the raw string info from self._raw_info() and formats it into Dictionary format.:
        Used to set self.info.

        Returns:
            dict. Dictionary is formatted as follows::
                >>> player_info = {
                ... 'is_suffix': True,
                ... 'recruiting': True,
                ... 'name': 'nriver',
                ... 'clan': 'Atlantis',
                ... 'title': 'the Liberator'
                ... }
            is_suffix (bool): If the player's title is a Suffix or not

            recruiting (bool): If the player's clan is set as Recruiting or not

            name (str): The player's name, passed as is when creating object Player

            clan (str): The player's clan name

            title (str): The player's current title
        """
        str_info = self._raw_info()
        info_list = []

        # str_info[36] = Start of json format in URL '{'
        for letter in str_info[36:]:
            info_list.append(letter)
            if letter == '}':
                break
        info_list = ''.join(info_list)
        info_dict = json.loads(info_list)
        info_dict['is_suffix'] = info_dict['isSuffix']
        del info_dict['isSuffix']
        return info_dict

    def set_runemetrics_info(self) -> None:
        """
        Sets Runemetrics info attributes for the Player.

        Attributes set:
            name (str): Case-sensitive player username.

            exp (int): Total Exp of the Player.

            combat_level (int): Total Combat Level of the Player.

            total_level (int): Total Skill Level of the Player.
        """
        # If user's runemetrics profile is private, self.name will be the same as passed when creating object.
        # Otherwise it will get the correct case-sensitive name from his runemetrics profile.
        # Some other info like Total exp, Combat level and Total level will be created as well.
        self.metrics_info = self.runemetrics_info()
        if self.metrics_info:
            if self.metrics_info.get('name'):
                self.name = self.metrics_info.get('name')
            self.exp = self.metrics_info.get('totalxp')
            self.combat_level = self.metrics_info.get('combatlevel')
            self.total_level = self.metrics_info.get('totalskill')
            self.quests_not_started = self.metrics_info.get('questsnotstarted')
            self.quests_started = self.metrics_info.get('questsstarted')
            self.quests_complete = self.metrics_info.get('questscomplete')
            self.skill_values = self.metrics_info.get('skillvalues')

    def runemetrics_info(self):
        """
        Gets player info from Runemetrics' API.
        Sets `private_profile` to True if his Runemetrics profile is found to be Private.

        Format::
            >>> {
            ... "magic":16216354,
            ... "questsstarted":5,
            ... "totalskill":2715,
            ... "questscomplete":198,
            ... "questsnotstarted":32,
            ... "totalxp":1037291112,
            ... "ranged":84195157,
            ... "activities":[
            ...
            ... ],
            ... # "skillvalues" has all skills and they are marked with their respective ID's, only 3 were shown here
            ... "skillvalues":[
            ...    {
            ...       "level":120,
            ...       "xp":1857550415,
            ...       "rank":12014,
            ...       "id":26
            ...    },
            ...    {
            ...       "level":99,
            ...       "xp":1202360390,
            ...       "rank":10370,
            ...       "id":6
            ...    },
            ...    {
            ...       "level":99,
            ...       "xp":841951573,
            ...       "rank":26941,
            ...       "id":3
            ...    },
            ... ],
            ... "name":"NRiver",
            ... "rank":"36,708",
            ... "melee":527931306,
            ... "combatlevel":138,
            ... "loggedIn":"false"
            ... }

        Returns:
            dict or bool. Runemetrics player info if the player profile is public, or False if it is Private.

        Raises:
            ConnectionError
                If connecting to the RS3 Runemetrics profile API was not possible.
        """
        user_name = self.name.replace(' ', '%20')
        info_url = f"https://apps.runescape.com/runemetrics/profile/profile?user={user_name}&activities=0"
        response = requests.get(info_url)
        self.runemetrics_status_code = response.status_code
        if response.status_code != 200:
            raise ConnectionError(
                f'Not able to connect to RS3 Runemetrics player profile API. Status code: {response.status_code}')
        json_info = response.json()
        if json_info.get('error') == 'PROFILE_PRIVATE':
            self.private_profile = True
            return None
        elif json_info.get('error') == 'NO_PROFILE':
            self.private_profile = False
            self.exists = False
            return None
        else:
            self.private_profile = False
            return json_info

    def skill(self, skill: str or int):
        """
        Gets information on a specific skill from the Player

        Usage::
            >>> # Example of usage of the method skill()
            >>> player = Player('NRiver')
            >>> player.skill('agility').level
            99
            >>> # Can pass skill names as well as id
            >>> # (8 = Woodcutting for example)
            >>> player.skill(8).exp
            14054178.6
            >>> player.skill('AtTaCk').rank
            68311

        Args:
            skill (str or int): Skill name or id to get info of

        Returns:
            Skill.
            Attributes of Skill
            --------------------
                exp (int): Total exp the Player has in that skill

                xp (int): Total exp the Player has in that skill

                level (int): The level the Player has in that skill

                rank (int): The Rank the Player has in that skill
        """
        skills_index = {
            0: 'attack',
            1: 'defence',
            2: 'strength',
            3: 'constitution',
            4: 'ranged',
            5: 'prayer',
            6: 'magic',
            7: 'cooking',
            8: 'woodcutting',
            9: 'fletching',
            10: 'fishing',
            11: 'firemaking',
            12: 'crafting',
            13: 'smithing',
            14: 'mining',
            15: 'herblore',
            16: 'agility',
            17: 'thieving',
            18: 'slayer',
            19: 'farming',
            20: 'runecrafting',
            21: 'hunter',
            22: 'construction',
            23: 'summoning',
            24: 'dungeoneering',
            25: 'divination',
            26: 'invention',
        }

        try:
            skill = int(skill)
            for _skill in self.skill_values:
                if _skill['id'] == skill:
                    return Skill(name=skills_index[skill], skill=_skill)
        except ValueError:
            skill = skill.lower()
            for index, value in skills_index.items():
                if skill == value:
                    skill = index
                    for _skill in self.skill_values:
                        if _skill['id'] == skill:
                            return Skill(name=skills_index[skill], skill=_skill)

    def __repr__(self):
        return f"Player(name={repr(self.name)}, clan={repr(self.clan)}, exists={self.exists})"


@dataclass
class Skill:
    """
    Represents a RuneScape 3 Skill

    Args:
        name (str): Name of the Skill

        skill (dict): Skill in dictionary format as it is in RuneScape's 3 API

    Attributes:
        exp (int): Total Exp the player has in the skill

        xp (int): Total Exp the player has in the skill

        level (int): Total Level the player has in the skill

        rank (int): Rank the player has in the skill

        id (int): The ID of the Skill
    """
    name: str
    skill: dict

    exp: int = None
    xp: int = None
    level: int = None
    rank: int = None
    id: int = None

    def __post_init__(self):
        self.exp = self.skill.get('xp') / 10
        self.xp = self.exp

        self.level = self.skill.get('level')
        self.rank = self.skill.get('rank')
        self.id = self.skill.get('id')

    def __str__(self):
        return self.name



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['rs3clans']

package_data = \
{'': ['*']}

install_requires = \
['requests>=2.21,<3.0']

extras_require = \
{':python_version >= "3.6" and python_version < "3.7"': ['dataclasses>=0.6,<0.7']}

setup_kwargs = {
    'name': 'rs3clans',
    'version': '1.1.1',
    'description': "A Python 3 module wrapper for RuneScape 3's API",
    'long_description': '# rs3clans.py\n[![PyPI](https://img.shields.io/pypi/v/rs3clans.svg)](https://pypi.org/project/rs3clans/) [![PyPI - Python Version](https://img.shields.io/pypi/pyversions/rs3clans.svg)](https://pypi.org/project/rs3clans/) [![Build Status](https://travis-ci.org/johnvictorfs/rs3clans.py.svg?branch=master)](https://travis-ci.org/johnvictorfs/rs3clans.py)\n\nA Python 3 module wrapper for RuneScape 3 API\n\n- [Change Log](CHANGELOG.md)\n\n***\n## Requirements:\n\n- [`Python 3.6+`](https://www.python.org/)\n\n- [`requests>=2.21.0`](http://docs.python-requests.org/en/master/)\n\n***\n\n## Setup with pip:\n\n```bash\n$ python3 -m pip install rs3clans\n```\n\n***\n\n## Usage:\n\n<details>\n<summary>\n> Players\n</summary>\n\n- Creating a Player object\n    - Always check if a player actually exists before doing anything with it\n```python\n>>> from rs3clans import players\n>>> player = players.Player(name=\'nriver\')\n>>> if player.exists:\n...     pass\n```\n\n- Whether the player exists or not\n```python\n>>> player.exists\nTrue\n```\n\n- Whether his Runemetrics Profile is Private or not\n```python\n>>> player.private_profile\nFalse\n```\n\n- You can also pass the argument runemetrics as `False` if you don\'t want their runemetrics info to be set\n    - This will make you unable to use some attributes from the Player class\n```python\n>>> player = players.Player(name=\'nriver\', runemetrics=False)\n```\n\n- Getting a player\'s name\n    - (if his Runemetrics Profile is private it will return the same name passed when creating object)\n```python\n>>> player.name\n\'NRiver\'\n```\n\n- Getting a player\'s total Exp (requires Public Runemetrics Profile)\n```python\n>>> player.exp\n1037291112\n```\n\n- Getting a player\'s Total Level (requires Public Runemetrics Profile)\n```python\n>>> player.total_level\n```\n\n- Getting a player\'s Combat Level (requires Public Runemetrics Profile)\n```python\n>>> player.combat_level\n138\n```\n\n- Quests information about a player (requires Public Runemetrics Profile)\n```python\n>>> player.quests_not_started\n32\n>>> player.quests_started\n5\n>>> player.quests_complete\n198\n```\n\n- Getting information on a specific skill of the player (requires Public Runemetrics Profile)\n```python\n>>> player.skill(\'agility\').level\n99\n```\n\n- Skill name is case-insensitive\n```python\n>>> player.skill(\'AtTaCk\').rank\n68311\n```\n\n- Can pass skill names as well as id\n    - (8 = Woodcutting for example)\n```python\n>>> player.skill(8).exp\n14054178.6\n```\n\n- Getting a player\'s title\n```python\n>>> player.title\n\'The Liberator\'\n```\n\n- Verifying if a player\'s title is a suffix or not\n```python\n>>> player.suffix\nTrue\n```\n\n- Getting a player\'s clan\n```python\n>>> player.clan\n\'Atlantis\'\n```\n</details>\n\n<details>\n<summary>\n> Clans\n</summary>\n\n- Creating a Clan object\n    - Always check if a clan actually exists before doing anything with it\n```python\n>>> from rs3clans import clans\n>>> try:\n...     clan = clans.Clan(\'Atlantis\')\n... except clans.ClanNotFoundError:\n...     print(\'Clan not found.\')\n```\n\n- Getting a clan\'s total Exp\n```python\n>>> clan.exp\n151349638333\n```\n\n- Getting information about a specific member in that clan\n    - Clan.member attribute (dict) (requires case-sensitive name)\n    - Clan.get_member() (method) (does not require case-sensitive name)\n    - Returns a ClanMember Object\n```python\n>>> # Case-sensitive\n>>> clan.member[\'NRiver\']\nClanMember(NRiver, Overseer, 1043065027)\n>>> clan.member[\'NRiver\'].rank\n\'Overseer\'\n```\n\n```python\n>>> # Case-insensitive\n>>> clan.get_member(\'nriver\')\nClanMember(NRiver, Overseer, 1043065027)\n>>> clan.get_member(\'nRiVeR\').rank\n\'Overseer\'\n```\n\n- Getting the number of players in a clan\n```python\n>>> clan.count\n499\n```\n\n- Getting the average Clan Exp per player in clan\n```python\n>>> clan.avg_exp\n303305888.44288576\n```\n\n- Iterate through a Clan\n```python\n>>> for member in clan:\n>>>     print(f"{member} - {member.name}")\nClanMember(Pedim, Owner, 1249520826) - Pedim\nClanMember(Acriano, Overseer, 1903276564) - Acriano\nClanMember(Cogu, Overseer, 1829449412) - Cogu\nClanMember(Black\xa0bullet, Overseer, 1100767386) - Black Bullet\nClanMember(NRiver, Overseer, 1090093362) - NRiver\nClanMember(Kurenaii, Overseer, 395850997) - Kurenaii\n...\n```\n\n</details>\n\n***\n\n## Contributing:\n\n- Guidelines:\n    - Follow Pep8 with the exception of `E501 (line too long)`\n\n- [Fork](https://github.com/johnvictorfs/rs3clans.py/fork) the repository\n\n- Install Dev dependencies\n    - With poetry\n        ```bash\n        $ python3 -m virtualenv .venv\n        $ poetry shell\n        $ poetry install\n        ```\n    - Without pipenv\n        ```bash\n        $ python3 -m virtualenv .venv\n        $ source .venv/bin/activate\n        $ pip install -r requirements-dev.txt\n        ```\n\n- Run the tests to make sure everything is ok\n    ```bash\n    $ pytest\n    ```\n\n- Make your changes to the code in the directory `rs3clans.py/rs3clans/`\n\n- Add necessary tests to the `rs3clans.py/tests/` directory ([`pytest docs`](https://docs.pytest.org/en/latest/))\n\n- Run the tests again\n    ```bash\n    $ pytest\n    ```\n\n- Also run `flake8` just to check if the code style is also fine\n    ```bash\n    $ flake8 --ignore=E501 rs3clans/\n    ```\n\n- If everything went fine then send a [Pull Request](https://github.com/johnvictorfs/rs3clans.py/pulls)\n\n***',
    'author': 'johnvictorfs',
    'author_email': 'johnvictorfs@gmail.com',
    'url': 'https://github.com/johnvictorfs/rs3clans.py',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



from dataclasses import dataclass
import csv

import requests


class ClanNotFoundError(Exception):
    """
    Error exception to be called when a clan is not found when trying to read the clan's URL using Rs3's Official API.

    Correct error handling when reading a clan::
        >>> import rs3clans
        >>> try:
        ...     clan_name = "oasdiuahsiubasiufbasuf"
        ...     clan = rs3clans.Clan(name=clan_name)
        ... except rs3clans.ClanNotFoundError:
        ...     print("Exception encountered!")
        Exception encountered!
    """

    def __init__(self, value):
        self.value = value

    def __str__(self):
        return repr(self.value)


@dataclass
class ClanMember:
    """
    Represents Member of a Clan

    Args:
        name (str): Name of the Clan Member
        rank (str): Rank of the Clan Member
        exp (int): Clan Exp of the Clan Member
    """
    name: str
    rank: str
    exp: int

    def __repr__(self) -> str:
        return f"ClanMember({self.name}, {self.rank}, {self.exp})"


@dataclass
class Clan:
    """Class with attributes of a Runescape Clan using Jagex's RS3 API

    Most of Runescape 3's API can be accessed at: http://runescape.wikia.com/wiki/Application_programming_interface

    ...

    Args:
        name (str): The name of the clan. Does not need to be case-sensitive.
    Kwargs:
        update_stats (bool): If True, updated stats for the Clan will be requested after creating the object and set.
        set_exp (bool): If True, the `exp` attribute will be set with the total Exp from the clan. False by default.
        set_dict (bool): If True, `member` attribute (dict) will be set with all members of the clan. True by default.

    Attributes:
        name (str): The name of the clan. Set when creating object.
        exp (int or None): The total Exp of the clan, or None if argument set_exp is False (False by default).
        member (dict or None): All info from members in the clan in dictionary format as below::
            >>> member_info = {
            ...     'player_1': {
            ...         'exp': 225231234,
            ...         'rank': 'Leader'
            ...     },
            ...     'player_2': {
            ...     'exp': 293123082,
            ...     'rank': 'Overseer'
            ...    },
            ... }
        or None if argument set_dict is False (True by default).
    count (int): The number of members in the Clan.
    avg_exp (int or None): The average clan exp per member in the Clan.

    Raises:
        ClanNotFoundError
            If an invalid clan name is passed when creating object Clan.
    """

    name: str
    update_stats: bool = True
    set_exp: bool = False
    set_dict: bool = True

    exp: int = None
    member: dict = None
    count: int = None
    avg_exp: float = None

    def __post_init__(self) -> None:
        if not self.name:
            raise TypeError("Clan name cannot be null.")
        if self.update_stats:
            self.update()

    def update(self, clan_list: list = None) -> None:
        """Updates all the Clan stats"""
        if not clan_list:
            clan_list = self.parse_clan_list(self.request_clan_list())

        if self.set_exp:
            self.exp = self._list_sum(clan_list)
        if self.set_dict:
            self.member = self.dict_lookup(clan_list)
            self.count = len(self.member)
        if self.set_exp and self.set_dict:
            self.avg_exp = self.exp / self.count

    @property
    def leader(self):
        """Finds the Clan leader in the clan member dict attribute"""
        for key, member in self.member.items():
            if member.rank == 'Owner':
                return member

    def request_clan_list(self) -> str:
        """Requests a Clan list in csv format from RS3's API from the Clan's Name"""

        clan_url = f'http://services.runescape.com/m=clan-hiscores/members_lite.ws?clanName={self.name}'

        with requests.Session() as session:
            download = session.get(clan_url)
            return download.content.decode('windows-1252')

    def parse_clan_list(self, decoded: str) -> list:
        """
        Parses a Clan List csv and returns it in a Python List

        Returns:
            list. All players information from a clan in list format like so::
                    >>> clan_list = [
                    ... ['Clanmate', ' Clan Rank', ' Total XP', ' Kills'],
                    ... ['Pedim', 'Owner', '739711654', '2'],
                    ... ['Tusoroxo', 'Deputy Owner', '1132958333', '0'],
                    ... ]
        """

        clan_list = list(csv.reader(decoded.splitlines(), delimiter=','))

        if clan_list[0][0] != "Clanmate":
            raise ClanNotFoundError(f"Couldn't find clan: {self.name}")

        for row in clan_list:
            # Replacing non-breaking spaces with actual spaces
            print(row[0])
            row[0] = row[0].replace(r"\xa0", " ")
        return clan_list

    @staticmethod
    def dict_lookup(clan_list: list) -> dict:
        """

        Used for getting all information available from a clan using Rs3's Clan API.

        Contrary to `list_lookup` it returns it as a Dictionary format instead.
        The dict format makes easier to find info specific to certain members of the Clan instead of looping over it.

        Args:
            clan_list (list): The clan list to make the dictionary for

        Returns:
            dict[str, ClanMember]. Information from all members in the clan in dictionary format for ease of access.
        """
        return {row[0]: ClanMember(row[0], row[1], int(row[2])) for row in clan_list[1:]}

    def get_member(self, name: str) -> ClanMember:
        """

        Used for searching information about a clan member by passing in its name case insensitively.

        Args:
            name (str): The name of the player to be searched in clan.

        Returns:
            ClanMember. ClanMember object of the found (if found) player.
        """
        for key, value in self.member.items():
            if key.lower() == name.lower():
                return value

    @staticmethod
    def _list_sum(clan_list: list) -> int:
        """

        Returns the Exp sum of a clan list.

        Returns:
            int. The total Exp of the Clan.
        """
        return sum(int(row[2]) for row in clan_list[1:])

    def __repr__(self) -> str:
        return f"Clan(name={repr(self.name)}, exp={self.exp}, avg_exp={self.avg_exp}, count={self.count})"

    def __iter__(self) -> ClanMember:
        """

        Iterates through `self.member` dict attribute, if `set_dict` was passed True when instancing `Clan` object.

        Yields:
            ClanMember. ClanMember object of each member in the clan::
                >>> import rs3clans
                >>> clan = rs3clans.Clan('atlantis')
                >>> for member in clan:
                ...     print(f"{member} - {member.name}")
                ...
                ClanMember(Pedim, Owner, 1249520826) - Pedim
                ClanMember(Acriano, Overseer, 1903276564) - Acriano
                ClanMember(Cogu, Overseer, 1829449412) - Cogu
                ClanMember(Black bullet, Overseer, 1100767386) - Black Bullet
                ClanMember(NRiver, Overseer, 1090093362) - NRiver
                ClanMember(Kurenaii, Overseer, 395850997) - Kurenaii
                (...)
        """
        if self.member:
            for key, value in self.member.items():
                yield value
